"""
    Works with source code reading.
"""

import tokenize
from typing import Dict, List, Union


def get_context_lines_from_source_code(
    filename: str, line_number: int, context_lines_count: int = 5
) -> Dict[str, Union[str, None, List[str]]]:
    """
    Returns context lines from source code file.
    """
    source_code_lines = _get_lines_from_source_code(filename=filename)
    bounds_start = max(0, line_number - context_lines_count - 1)
    bounds_end = min(line_number + 1 + context_lines_count, len(source_code_lines))

    strip_line = lambda line: line.strip("\r\n").replace("    ", "\t")
    context_pre, context_target, context_post = [], None, []
    try:
        context_pre = [
            strip_line(line)
            for line in source_code_lines[bounds_start : line_number - 1]
        ]
        context_target = strip_line(source_code_lines[line_number - 1])
        context_post = [
            strip_line(line) for line in source_code_lines[line_number:bounds_end]
        ]
    except IndexError:
        # File was changed?
        pass
    return {"pre": context_pre, "target": context_target, "post": context_post}


def _get_lines_from_source_code(filename: str) -> List[str]:
    """
    Returns lines of the code from the source code filename.
    """
    try:
        with tokenize.open(filename=filename) as source_file:
            return source_file.readlines()
    except (OSError, IOError):
        return []



"""
    Utils stuff.
"""

import sys
import platform

from typing import Any, Dict

from gatey_sdk.consts import SDK_INFORMATION_DICT
from gatey_sdk.consts import (
    RUNTIME_NAME,
)


def remove_trailing_slash(url: str) -> str:
    """
    Removes trailing slash from a URL.
    Example: `http://example.com/` will become `http://example.com` (No trailing slash)

    :param url: The URL to remove trailing slash.
    """
    if not isinstance(url, str):
        raise TypeError("URL must be a string!")

    if url.endswith("/"):
        url = url[:-1]

    return url


def get_additional_event_tags(
    include_platform_info: bool = True,
    include_runtime_info: bool = True,
    include_sdk_info: bool = True,
) -> Dict[str, Any]:
    """
    Returns additional event dictionary for tags with event information such as SDK information, platform information etc.
    """
    additional_event_tags = dict()
    if include_sdk_info:
        additional_event_tags.update(SDK_INFORMATION_DICT)
    if include_platform_info:
        additional_event_tags.update(get_platform_event_tags())
    if include_runtime_info:
        additional_event_tags.update(get_runtime_event_tags())
    return additional_event_tags


def get_platform_event_tags() -> Dict[str, Any]:
    """
    Returns platform information for event data tags.
    """

    platform_os = platform.system()
    platform_network_name = platform.node()
    platform_event_data_tags = {
        "os": platform_os,
        "os.node": platform_network_name,
        "os.version": platform.version(),  # For major there is `platform.release()`
        "os.bits": platform.architecture()[0],
    }

    return platform_event_data_tags


def get_runtime_event_tags() -> Dict:
    """
    Returns runtime information event data tags.
    """
    runtime_name = RUNTIME_NAME
    runtime_version = sys.version_info
    runtime_version = f"{runtime_version[0]}.{runtime_version[1]}.{runtime_version[2]}-{runtime_version[3]}-{runtime_version[4]}"
    return {
        "runtime.name": runtime_name,
        "runtime.ver": runtime_version,
        "runtime.impl": platform.python_implementation(),
    }



"""
    Stuff to work with tracebacks.
"""

from typing import Dict, List
from types import TracebackType
from gatey_sdk.internal.source import get_context_lines_from_source_code


def get_trace_from_traceback(
    traceback: TracebackType,
    include_code_context: bool = True,
    code_context_lines_count: int = 5,
    code_context_only_for_tail: bool = True,
) -> List[Dict]:
    """
    Returns trace from the given traceback.
    """

    trace = []

    while traceback is not None:
        # Iterating over traceback with `tb_next`
        frame = traceback.tb_frame
        frame_code = getattr(frame, "f_code", None)
        filename, function = None, None
        if frame_code:
            filename = frame.f_code.co_filename
            function = frame.f_code.co_name

        line_number = traceback.tb_lineno
        trace_element = {
            "filename": filename,
            "name": function or "<unknown>",
            "line": line_number,
            "module": frame.f_globals.get("__name__", None),
        }

        if include_code_context and not code_context_only_for_tail:
            trace_element |= {
                "context": get_context_lines_from_source_code(
                    filename=filename,
                    line_number=line_number,
                    context_lines_count=code_context_lines_count,
                )
            }

        trace.append(trace_element)
        traceback = traceback.tb_next

    if include_code_context and code_context_only_for_tail:
        tail_trace = trace[-1]
        tail_trace["context"] = get_context_lines_from_source_code(
            filename=tail_trace["filename"],
            line_number=tail_trace["line"],
            context_lines_count=5,
        )

    return trace


def get_variables_from_traceback(
    traceback: TracebackType, *, _always_skip: bool = False
) -> Dict:
    """
    Returns local and global variables from the given traceback.
    """

    traceback_variables_locals = {}
    traceback_variables_globals = {}

    if traceback and not _always_skip:
        traceback_variables_locals = traceback.tb_frame.f_locals
        traceback_variables_globals = traceback.tb_frame.f_globals

    # Stringify variable values.
    traceback_variables_locals = {
        key: str(value) for key, value in traceback_variables_locals.items()
    }
    traceback_variables_globals = {
        key: str(value) for key, value in traceback_variables_globals.items()
    }

    return {
        "locals": traceback_variables_locals,
        "globals": traceback_variables_globals,
    }



"""
    Flask integration(s).
"""

from typing import Callable, Dict, Any, Optional
from werkzeug.wrappers import Request, Response
from flask import abort
from gatey_sdk.client import Client

# Type aliases for callables.
HookCallable = Callable[["GateyFlaskMiddleware", Dict, Callable], None]
ClientGetterCallable = Callable[[], Client]


class GateyFlaskMiddleware:
    """Gatey SDK Starlette middleware."""

    # Requirements.
    flask_app: Callable[[Dict, Callable], Any]
    gatey_client: Client

    # Gatey options.
    capture_exception_options: Dict[str, Any] = {"include_default_tags": True}
    pre_capture_hook: HookCallable
    post_capture_hook: HookCallable
    client_getter: ClientGetterCallable
    capture_requests_info: bool = False
    capture_requests_info_additional_tags: Dict[str, str] = dict()

    def __init__(
        self,
        app,
        client: Client | None = None,
        *,
        capture_requests_info: bool = False,
        client_getter: Optional[ClientGetterCallable] = None,
        capture_exception_options: Optional[Dict[str, Any]] = None,
        pre_capture_hook: Optional[HookCallable] = None,
        post_capture_hook: Optional[HookCallable] = None,
        on_request_hook: Optional[HookCallable] = None,
        capture_requests_info_additional_tags: Optional[Dict[str, str]] = None,
    ) -> None:
        self.flask_app = app
        self.gatey_client = client  # Redefined below by `client_getter`.

        self.capture_requests_info = capture_requests_info
        self.capture_exception_options = (
            capture_exception_options
            if capture_exception_options
            else self.capture_exception_options
        )
        self.capture_requests_info_additional_tags = (
            capture_requests_info_additional_tags
            if capture_requests_info_additional_tags
            else dict()
        )
        self.client_getter = (
            client_getter if client_getter else self._default_client_getter
        )

        # Hooks.
        hooks = {
            "pre_capture_hook": pre_capture_hook,
            "post_capture_hook": post_capture_hook,
            "on_request_hook": on_request_hook,
        }
        for name, hook in hooks.items():
            setattr(self, name, hook or self._default_void_hook)

        self.gatey_client = self.client_getter()
        if not isinstance(self.gatey_client, Client):
            raise ValueError(
                "Gatey client is invalid! Please review `client` param or review your client getter!"
            )

    def __call__(self, environ: Dict, start_response: Callable) -> None:
        """
        Middleware itself (handle request).
        """
        return self._execute_app_wrapped(environ, start_response)

    def _execute_app_wrapped(self, environ: Dict, start_response: Callable) -> Any:
        """
        Executes app wrapped with middleware.
        """
        app_args = [environ, start_response]

        if self.on_request_hook:
            self.on_request_hook(self, *app_args)

        if self.capture_requests_info:
            self._capture_request_info(environ=environ)

        try:
            return self.flask_app(*app_args)
        except Exception as _flask_app_exception:
            client = self.client_getter()
            if client and isinstance(client, Client):
                self.pre_capture_hook(self, *app_args)

                capture_options = self.capture_exception_options.copy()
                if "tags" not in capture_options:
                    capture_options["tags"] = self._get_request_tags_from_environ(
                        environ=environ
                    )

                client.capture_exception(_flask_app_exception, **capture_options)

                self.post_capture_hook(self, *app_args)
            abort(500)

    def _get_request_tags_from_environ(self, environ: Dict) -> Dict[str, str]:
        """
        Returns tags for request from request environ.
        """
        request = Request(environ)
        return {
            "query": request.query_string.decode("utf-8"),
            "path": request.root_path + request.path,
            "scheme": request.scheme,
            "method": request.method,
            "client_host": request.remote_addr,
            "server_host": ":".join(map(str, request.server)),
        }

    def _capture_request_info(self, environ: Dict) -> None:
        """
        Captures request info as message to the client.
        """
        client = self.client_getter()
        if not client:
            return

        tags = self._get_request_tags_from_environ(environ=environ)
        message = f"{tags['method']} '{tags['path']}'"
        message = message if message != " ''" else "Request was handled."
        tags.update(self.capture_requests_info_additional_tags)

        client.capture_message(
            message,
            level="debug",
            tags=tags,
        )

    def _default_client_getter(self) -> Client:
        """
        Default getter for Gatey client if none is specified.
        """
        return self.gatey_client

    def _default_void_hook(*_) -> None:
        """
        Default hook for pre/post capture, just does nohing.
        """
        return None



"""
    Florgon Gatey SDK.

    SDK Library for Florgon Gatey API.
    Provides interface for working with Florgon Gatey.

    Florgon Gatey is a error logging service.
    Library provides interface for working with exceptions, catching them,
    Working with message / events and all another analytics stuff,
    Also there is base API interface, allowing to work with native API, not base gateway end-user API.

    Author: Florgon Solutions.
    Gatey website: https://gatey.florgon.space/
    Gatey API endpoint: https://api.florgon.space/gatey/
    Gatey developer documentation: https://gatey-sdk-py.readthedocs.io/

    If you have any questions please reach out us at:
    - support@florgon.space

    Main SDK maintainer:
    - Kirill Zhosul (@kirillzhosul)
    - kirillzhosul@florgon.space
    - https://github.com/kirillzhosul
"""

# Library specific information.
from gatey_sdk.__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)

# Base API.
from gatey_sdk.client import Client
from gatey_sdk.response import Response

# Internal exceptions.
from gatey_sdk.exceptions import GateyApiError, GateyTransportError

# Additional API.
from gatey_sdk.api import Api
from gatey_sdk.response import Response
from gatey_sdk.transport import (
    BaseTransport,
    HttpTransport,
    FuncTransport,
    VoidTransport,
    PrintTransport,
)


__all__ = [
    "Client",
    "Response",
    "Api",
    "BaseTransport",
    "HttpTransport",
    "FuncTransport",
    "VoidTransport",
    "PrintTransport",
    "GateyApiError",
    "GateyTransportError",
]



"""
    Transport classes, that handles sending events from Gatey.
"""
import json
from typing import Callable, Any, Union, Optional, Dict
from gatey_sdk.exceptions import (
    GateyError,
    GateyTransportError,
    GateyTransportImproperlyConfiguredError,
)
from gatey_sdk.api import Api
from gatey_sdk.auth import Auth


def _transport_base_sender_wrapper(func):
    """
    Wrapper for transports send event methods that converts result to success state.
    """

    def wrapper(*args, **kwargs) -> bool:
        fail_fast = kwargs.pop("__fail_fast", False)
        try:
            func(*args, **kwargs)
        except GateyError as internal_exception:
            if fail_fast:
                raise internal_exception
            return False
        else:
            return True

    return wrapper


class BaseTransport:
    """
    Base transport class. Cannot be used as transport.
    Abstract class for implementing transport classes.
    """

    def __init__(self):
        pass

    @_transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Handles transport event callback (handle event sending).
        Should be inherited from BaseTransport and implemented in transports.
        """
        raise NotImplementedError()


class HttpTransport(BaseTransport):
    """
    HTTP Transport. Sends event to the Gatey Server when event sends.
    """

    # Allowed aggreation / composition..
    _api_provider: Api = None
    _auth_provider: Optional[Auth] = None

    def __init__(self, api: Optional[Api] = None, auth: Optional[Auth] = None):
        """
        :param api: Api provider.
        :param auth: Authentication provider.
        """

        BaseTransport.__init__(self)
        self._auth_provider = auth if auth else Auth()
        self._api_provider = api if api else Api(self._auth_provider)
        self._check_improperly_configured()

    @_transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        api_params = self._api_params_from_event_dict(event_dict=event_dict)
        self._api_provider.method("event.capture", send_project_auth=True, **api_params)

    def _api_params_from_event_dict(self, event_dict: Dict):
        api_params = {
            "level": event_dict["level"],
        }

        event_params = ["exception", "message", "tags"]

        for param in event_params:
            if param in event_dict:
                api_params[param] = json.dumps(event_dict[param])

        return api_params

    def _check_improperly_configured(self):
        """
        Raises error if auth provider improperly configured for sending event.
        """
        if self._auth_provider.project_id is None:
            raise GateyTransportImproperlyConfiguredError(
                "HttpTransport improperly configured! No project id found in auth provider!"
            )
        if (
            self._auth_provider.server_secret is None
            and self._auth_provider.client_secret is None
        ):
            raise GateyTransportImproperlyConfiguredError(
                "HttpTransport improperly configured! No client / server secret found in auth provider!"
            )


class FuncTransport(BaseTransport):
    """
    Function transport. Calls your function when event sends.
    """

    def __init__(self, func: Callable):
        """
        :param function: Function to call when event sends.
        """
        BaseTransport.__init__(self)
        self._function = func

    @_transport_base_sender_wrapper
    def send_event(
        self, event_dict: Dict, skip_to_internal_exception: bool = False
    ) -> None:
        """
        Handles transport event callback (handle event sending).
        Function transport just takes event and passed it raw to function call.
        """
        if skip_to_internal_exception:
            try:
                self._function(event_dict)
            except Exception as _:
                raise GateyTransportError(
                    f"Unable to handle event send with Function transport (FuncTransport)."
                )
            return
        self._function(event_dict)


class VoidTransport(BaseTransport):
    """
    Void transport. Does nothing, used as test environment.
    """

    @_transport_base_sender_wrapper
    def send_event(self, *args, **kwargs) -> None:
        """
        Handles transport event callback (handle event sending).
        Does nothing.
        """
        pass


class PrintTransport(BaseTransport):
    """
    Print transport. Prints event data, used ONLY as test environment.
    """

    def __init__(
        self,
        indent: int | str | None = 2,
        prepare_event: Callable[[Dict], Dict] | None = None,
        print_function: Callable[[Dict], Any] | None = None,
    ):
        """
        :param function: Function to call when event sends.
        """
        BaseTransport.__init__(self)
        self._indent = indent
        self._prepare_event = prepare_event if prepare_event else lambda e: e
        self._print_function = print_function if print_function else print

    @_transport_base_sender_wrapper
    def send_event(self, event_dict: Dict) -> None:
        """
        Handles transport event callback (handle event sending).
        Print event data.
        """
        dump_data = (
            self._prepare_event(event_dict) if self._prepare_event else event_dict
        )
        print(
            json.dumps(
                dump_data,
                indent=self._indent,
                sort_keys=True,
            )
        )


def build_transport_instance(
    transport_argument: Any = None,
    api: Optional[Api] = None,
    auth: Optional[Auth] = None,
) -> Union[BaseTransport, None]:
    """
    Builds transport instance by transport argument.
    """
    transport_class = None

    if transport_argument is None:
        # If nothing is passed, should be default http transport type.
        return HttpTransport(api=api, auth=auth)

    if isinstance(transport_argument, type) and issubclass(
        transport_argument, BaseTransport
    ):
        # Passed subclass (type) of BaseTransport as transport.
        # Should be instantiated as cls.
        transport_class = transport_argument
        if transport_class in (VoidTransport, PrintTransport):
            return transport_class()
        try:
            return transport_class(api=api, auth=auth)
        except TypeError as _transport_params_error:
            raise GateyTransportImproperlyConfiguredError(
                "Failed to build transport instance. Please instantiate before or except your transport to handle `api`, `auth` params in constructor!"
            ) from _transport_params_error

    if isinstance(transport_argument, BaseTransport):
        # Passed already constructed transport, should do nothing.
        return transport_argument

    if callable(transport_argument):
        # Passed callable (function) as transport.
        # Should be Function transport, as it handles raw function call.
        return FuncTransport(function=transport_argument)

    # Unable to instantiate transport instance.
    raise GateyTransportImproperlyConfiguredError(
        "Failed to build transport instance. Please pass valid transport argument!"
    )



"""
    Library specific information.
"""

__title__ = "gatey-sdk"
__description__ = "Python client for Gatey (https://gatey.florgon.space)"
__url__ = "https://github.com/florgon/gatey-sdk-py"
__version__ = "0.0.7"
__author__ = "Florgon Team and Contributors"
__author_email__ = "support@florgon.space"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Florgon Solutions"
